# coding=utf-8

from requests import HTTPError
from ...base import BitbucketServerBase
from ...common.permissions import Groups, Users


class Repositories(BitbucketServerBase):
    def __init__(self, url, *args, **kwargs):
        super(Repositories, self).__init__(url, *args, **kwargs)

    def __get_object(self, data):
        return Repository(data, **self._new_session_args)

    def create(self, name):
        """
        Creates a new repository with the given name.

        See https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp174

        :param name: string: The name of the project.

        :return: The created project object
        """
        return self.__get_object(self.post(None, data={"name": name}))

    def each(self):
        """
        Get all repositories.

        See https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp175

        :return: A generator for the Repository objects
        """
        for repository in self._get_paged(None):
            yield self.__get_object(repository)

    def get(self, repository, by="slug"):
        """
        Returns the requested repository.

        :param repository: string: The requested repository.
        :param by: string (default is "slug"): How to interpret project, can be 'slug' or 'name'.

        :return: The requested Repository object
        """
        if by == "slug":
            return self.__get_object(super(Repositories, self).get(repository))
        elif by == "name":
            for r in self.each():
                print(r.name)
                if r.name == repository:
                    return r
        else:
            ValueError("Unknown value '{}' for argument [by], expected 'slug' or 'name'".format(by))

        raise Exception("Unknown repository {} '{}'".format(by, repository))

    def exists(self, repository, by="slug"):
        """
        Check if repository exist.

        :param repository: string: The requested repository.
        :param by: string (default is "slug"): How to interpret repository, can be 'slug' or 'name'.

        :return: True if the repository exists
        """
        exists = False
        try:
            self.get(repository, by)
            exists = True
        except HTTPError as e:
            if e.response.status_code in (401, 404):
                pass
        except Exception as e:
            if not str(e) == "Unknown project {} '{}'".format(by, repository):
                raise e
        return exists


class Repository(BitbucketServerBase):
    def __init__(self, data, *args, **kwargs):
        super(Repository, self).__init__(None, *args, data=data, **kwargs)
        self.__groups = Groups(self._sub_url("permissions/groups"), "REPO", **self._new_session_args)
        self.__users = Users(self._sub_url("permissions/users"), "REPO", **self._new_session_args)

    def __get_object(self, data):
        return Repository(data, **self._new_session_args)

    def update(self, **kwargs):
        """
        Update the repository properties. Fields not present in the request body are ignored.

        :param kwargs: dict: The data to update.

        :return: The updated repository

        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp180
        """
        return self._update_data(self.put(None, data=kwargs))

    def delete(self):
        """
        Delete the repostory.

        :return: The response on success

        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp177
        """
        return super(Repository, self).delete(None)

    @property
    def id(self):
        """The repository identifier"""
        return self.get_data("id")

    @property
    def name(self):
        """The repository name"""
        return self.get_data("name")

    @name.setter
    def name(self, name):
        """Setter for the repository name"""
        return self.update(name=name)

    @property
    def slug(self):
        """The repository slug"""
        return self.get_data("slug")

    @property
    def description(self):
        """The repository description"""
        return self.get_data("description")

    @description.setter
    def description(self, description):
        """Setter for the repository description"""
        return self.update(description=description)

    @property
    def public(self):
        """The repository public flag"""
        return self.get_data("public")

    @public.setter
    def public(self, public):
        """Setter for the repository public flag"""
        return self.update(public=public)

    @property
    def forkable(self):
        """The repository forkable flag"""
        return self.get_data("forkable")

    @forkable.setter
    def forkable(self, forkable):
        """Setter for the repository forkable flag"""
        return self.update(forkable=forkable)

    def contributing(self, at=None, markup=None):
        """
        Get the contributing guidelines.

        :param at: string: Optional, the commit to get the contributing guideline from.
        :param markup: boolean: Optional, If set to true, the rendered content is returned as HTML.

        :return: The text content of the contributing guidlines.

        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp183
        """
        params = {}
        if at is not None:
            params["at"] = at
        if markup is True:
            params["markup"] = markup

        return self.get("contributing", params=params)

    def license(self, at=None, markup=None):
        """
        Get the license file.

        :param at: string: Optional, the commit to get the license file from.
        :param markup: boolean: Optional, If set to true, the rendered content is returned as HTML.

        :return: The text content of the license file.

        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp191
        """
        params = {}
        if at is not None:
            params["at"] = at
        if markup is True:
            params["markup"] = markup

        return self.get("license", params=params)

    def readme(self, at=None, markup=None):
        """
        Get the readme file.

        :param at: string: Optional, the commit to get the readme file from.
        :param markup: boolean: Optional, If set to true, the rendered content is returned as HTML.

        :return: The text content of the readme file.

        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp194
        """
        params = {}
        if at is not None:
            params["at"] = at
        if markup is True:
            params["markup"] = markup

        return self.get("readme", params=params)

    @property
    def default_branch(self):
        """
        Get the default branch.

        :return: The default branch (displayId without refs/heads/).

        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp185
        """
        return self.get("default-branch")["displayId"]

    @default_branch.setter
    def default_branch(self, branch):
        """
        Set the default branch.

        :param id: string: The default branch to set (without refs/heads/).

        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp186
        """
        self.put("default-branch", data={"id": "refs/heads/{}".format(branch)})

    def forks(self):
        """
        Get all forks.

        :return: A generator object for the forks.

        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp188
        """
        for fork in self._get_paged("forks"):
            yield fork

    def related(self):
        """
        Get all related repositories.

        :return: A generator for the related repository objects

        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp195
        """
        for repository in self._get_paged("related"):
            yield self.__get_object(repository)

    @property
    def groups(self):
        """
        Property to access the project groups:
        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp279
        """
        return self.__groups

    @property
    def users(self):
        """
        Property to access the project groups
        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp285
        """
        return self.__users

    def download_archive(self, dest_fd, at=None, filename=None, format=None, path=None, prefix=None, chunk_size=128):
        """
        Downloads a repository archive.
        Note that the data is written to the specified file-like object,
        rather than simply being returned.
        For further information visit:
           https://docs.atlassian.com/bitbucket-server/rest/7.13.0/bitbucket-rest.html#idp199
        :param dest_fd: a file-like object to which the archive will be written
        :param at: string: Optional, the commit to download an archive of; if not supplied, an archive of the default branch is downloaded
        :param filename: string: Optional, a filename to include the "Content-Disposition" header
        :param format: string: Optional, the format to stream the archive in; must be one of: zip, tar, tar.gz or tgz. If not specified, then the archive will be in zip format.
        :param paths: string: Optional, path to include in the streamed archive
        :param prefix: string: Optional, a prefix to apply to all entries in the streamed archive; if the supplied prefix does not end with a trailing /, one will be added automatically
        :param chunk_size: int: Optional, download chunk size. Defeault is 128
        """
        params = {}
        if at is not None:
            params["at"] = at
        if filename is not None:
            params["filename"] = filename
        if format is not None:
            params["format"] = format
        if path is not None:
            params["path"] = path
        if prefix is not None:
            params["prefix"] = prefix
        headers = {"Accept": "*/*"}
        response = self.get("archive", params=params, headers=headers, advanced_mode=True)
        for chunk in response.iter_content(chunk_size=chunk_size):
            dest_fd.write(chunk)



# coding=utf-8

from .base import BitbucketServerBase


class GlobalPermissions(BitbucketServerBase):
    LICENSED_USER = "LICENSED_USER"
    PROJECT_CREATE = "PROJECT_CREATE"
    ADMIN = "ADMIN"
    SYS_ADMIN = "SYS_ADMIN"

    def __init__(self, url, *args, **kwargs):
        super(GlobalPermissions, self).__init__(url, *args, **kwargs)

    def licensed_user(self, name):
        """
        Add the licensed user permission for a group/user.
        """
        return self.add(name, self.__permission(self.LICENSED_USER))

    def project_create(self, name):
        """
        Add the project create permission for a group/user.
        """
        return self.add(name, self.__permission(self.PROJECT_CREATE))

    def admin(self, name):
        """
        Add the admin permission for a group/user.
        """
        return self.add(name, self.__permission(self.ADMIN))

    def sys_admin(self, name):
        """
        Add the sys admin permission for a group/user.
        """
        return self.add(name, self.__permission(self.SYS_ADMIN))

    def add(self, name, permission):
        """
        Add the permission for a group/user.

        :param name: string: The names of the groups/users
        :param permission: string: The permission to grant.

        API docs:
        - For groups see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp64
        - For users see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp70
        """
        self.put(name, permission)
        return

    def each(self, filter=None):
        """
        Get all groups/users.

        :params filter: string: If specified only group/user names containing the supplied string will be returned

        :return: A generator for the group/user permission objects

        API docs:
        - For groups see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp63
        - For users see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp69
        """
        params = {}
        if filter is not None:
            params["filter"] = filter
        for entry in self._get_paged(None, params=params):
            entry = self._get_object(entry)
            entry.url = self.url
            yield entry

    def each_none(self, filter=None):
        """
        Get all not granted groups/users.

        :params filter: string: If specified only group/user names containing the supplied string will be returned

        :return: A generator for the group/user permission objects

        API docs:
        - For groups see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp67
        - For users see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp73
        """
        params = {}
        if filter is not None:
            params["filter"] = filter
        for entry in self._get_paged("none", params=params):
            yield self._get_object(entry)

    def get(self, name):
        """
        Returns the requested group/user

        :param name: string: The requested element name.

        :return: The requested group/user object
        """
        for entry in self.each(filter=name):
            if entry.name == name:
                return entry

        raise Exception("Unknown group/user '{}'".format(name))


class Groups(GlobalPermissions):
    def __init__(self, url, *args, **kwargs):
        super(Groups, self).__init__(url, *args, **kwargs)

    def _get_object(self, data):
        return Group(data, **self._new_session_args)


class Users(GlobalPermissions):
    def __init__(self, url, *args, **kwargs):
        super(Users, self).__init__(url, *args, **kwargs)

    def _get_object(self, data):
        return User(data, **self._new_session_args)


class PermissionBase(BitbucketServerBase):
    @property
    def permission(self):
        if self.url is None:
            raise NotImplementedError("Permission not implemented for this object type.")
        return self.get_data("permission")

    @property
    def is_licensed_user(self):
        return True if self.permission == GlobalPermissions.LICENSED_USER else False

    @property
    def is_project_create(self):
        return True if self.permission == GlobalPermissions.PROJECT_CREATE else False

    @property
    def is_admin(self):
        return True if self.permission == GlobalPermissions.ADMIN else False

    @property
    def is_sys_admin(self):
        return True if self.permission == GlobalPermissions.SYS_ADMIN else False

    def delete(self):
        """
        Delete the permission.

        :return: The response on success
        """
        if self.url is None:
            raise NotImplementedError("Delete not implemented for this object type.")
        return super(PermissionBase, self).delete(None, params={"name": self.name})


class Group(PermissionBase):
    def __init__(self, data, *args, **kwargs):
        super(Group, self).__init__(None, *args, data=data, **kwargs)

    @property
    def name(self):
        if self.url is None:
            return self.get_data("name")
        return self.get_data("group")["name"]


class User(PermissionBase):
    def __init__(self, data, *args, **kwargs):
        super(User, self).__init__(None, *args, data=data, **kwargs)

    def __userdata(self, key):
        if self.url is None:
            return self.get_data(key)
        return self.get_data("user")[key]

    @property
    def name(self):
        return self.__userdata("name")

    @property
    def email(self):
        return self.__userdata("emailAddress")

    @property
    def displayname(self):
        return self.__userdata("displayName")

    @property
    def active(self):
        return self.__userdata("active")

    @property
    def slug(self):
        return self.__userdata("slug")

    @property
    def id(self):
        return self.__userdata("id")



# coding=utf-8

from ..base import BitbucketServerBase


class Permissions(BitbucketServerBase):
    ADMIN = "ADMIN"
    WRITE = "WRITE"
    READ = "READ"

    def __init__(self, url, permission_prefix, *args, **kwargs):
        self.__permission_prefix = permission_prefix
        super(Permissions, self).__init__(url, *args, **kwargs)

    def __permission(self, permission):
        """Internal function to get the permission for the put request."""
        return "{}_{}".format(self.__permission_prefix, permission)

    def admin(self, name):
        """
        Add the admin permission for a group/user.
        """
        return self.add(name, self.__permission(self.ADMIN))

    def write(self, name):
        """
        Add the write permission for a group/user.
        """
        return self.add(name, self.__permission(self.WRITE))

    def read(self, name):
        """
        Add the read permission for a group/user.
        """
        return self.add(name, self.__permission(self.READ))

    def add(self, name, permission):
        """
        Add the permission for a group/user.

        :param name: string: The names of the groups/users
        :param permission: string: The permission to grant.

        API docs:
        - For project groups see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp160
        - For project users see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp166
        - For repository groups see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp282
        - For repository users see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp288
        """
        self.put(name, permission)
        return

    def each(self, filter=None):
        """
        Get all groups/users.

        :params filter: string: If specified only group/user names containing the supplied string will be returned

        :return: A generator for the group/user permission objects

        API docs:
        - For project groups see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp159
        - For project users see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp165
        - For repository groupss see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp280
        - For repository users see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp286
        """
        params = {}
        if filter is not None:
            params["filter"] = filter
        for entry in self._get_paged(None, params=params):
            entry = self._get_object(entry)
            entry.url = self.url
            yield entry

    def each_none(self, filter=None):
        """
        Get all not granted groups/users.

        :params filter: string: If specified only group/user names containing the supplied string will be returned

        :return: A generator for the group/user permission objects

        API docs:
        - For project groups see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp163
        - For project users see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp169
        - For repository groups see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp284
        - For repository users see https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp290
        """
        params = {}
        if filter is not None:
            params["filter"] = filter
        for entry in self._get_paged("none", params=params):
            yield self._get_object(entry)

    def get(self, name):
        """
        Returns the requested group/user

        :param name: string: The requested element name.

        :return: The requested group/user object
        """
        for entry in self.each(filter=name):
            if entry.name == name:
                return entry

        raise Exception("Unknown group/user '{}'".format(name))


class Groups(Permissions):
    def __init__(self, url, permission_prefix, *args, **kwargs):
        super(Groups, self).__init__(url, permission_prefix, *args, **kwargs)

    def _get_object(self, data):
        return Group(data, **self._new_session_args)


class Users(Permissions):
    def __init__(self, url, permission_prefix, *args, **kwargs):
        super(Users, self).__init__(url, permission_prefix, *args, **kwargs)

    def _get_object(self, data):
        return User(data, **self._new_session_args)


class PermissionBase(BitbucketServerBase):
    def delete(self):
        """
        Delete the permission.

        :return: The response on success
        """
        if self.url is None:
            raise NotImplementedError("Delete not implemented for this object type.")
        return super(BitbucketServerBase, self).delete(None, params={"name": self.name})

    @property
    def permission(self):
        if self.url is None:
            raise NotImplementedError("Pemission not implemented for this object type.")
        return self.get_data("permission")

    @property
    def is_admin(self):
        """True if group/user is admin"""
        return True if self.permission == Permissions.ADMIN else False

    @property
    def is_write(self):
        """True if group/user has write permission"""
        return True if self.permission == Permissions.WRITE else False

    @property
    def is_read(self):
        """True if group/user has read premission"""
        return True if self.permission == Permissions.READ else False

    @property
    def can_write(self):
        """True if group/user can write"""
        return True if self.permission in (Permissions.ADMIN, Permissions.WRITE) else False


class Group(PermissionBase):
    def __init__(self, data, *args, **kwargs):
        super(Group, self).__init__(None, *args, data=data, **kwargs)

    @property
    def name(self):
        """The name of the group"""
        if self.url is None:
            return self.get_data("name")
        return self.get_data("group")["name"]


class User(PermissionBase):
    def __init__(self, data, *args, **kwargs):
        super(User, self).__init__(None, *args, data=data, **kwargs)

    def __userdata(self, key):
        """
        Internal getter for the user data.

        :param key: string: The user data to get

        :return: The requested user data
        """
        if self.url is None:
            return self.get_data(key)
        return self.get_data("user")[key]

    @property
    def name(self):
        """The name of the user"""
        return self.__userdata("name")

    @property
    def email(self):
        """The email of the user"""
        return self.__userdata("emailAddress")

    @property
    def displayname(self):
        """The displayname of the user"""
        return self.__userdata("displayName")

    @property
    def active(self):
        """The active flag of the user"""
        return self.__userdata("active")

    @property
    def slug(self):
        """The slug of the user"""
        return self.__userdata("slug")

    @property
    def id(self):
        """The id of the user"""
        return self.__userdata("id")



# coding=utf-8

from requests import HTTPError
from ..base import BitbucketCloudBase

from ..repositories import ProjectRepositories


class Projects(BitbucketCloudBase):
    def __init__(self, url, *args, **kwargs):
        super(Projects, self).__init__(url, *args, **kwargs)

    def __get_object(self, data):
        return Project(data, **self._new_session_args)

    def create(self, name, key, description, is_private=True, avatar=None):
        """
        Creates a new project with the given values

        Note that the avatar has to be embedded as either a data-url or a URL to an external image as shown in
        the examples below:

            w.projects.create( "Mars Project", "MARS", "Software for colonizing mars.",
                avatar="data:image/gif;base64,R0lGODlhEAAQAMQAAORHHOVSKudfOulrSOp3WOyDZu6QdvCchPGolfO0o/...",
                is_private=False
            )

            w.projects.create( "Mars Project", "MARS", "Software for colonizing mars.",
                avatar="http://i.imgur.com/72tRx4w.gif",
                is_private=False
            )

        :param name: string:                       The name of the project.
        :param key: string:                        The key of the project.
        :param description: string:                The description of the project.
        :param is_private: bool (default is True): True if it is a private project.
        :param avatar: string (default is None):   The avatar of the project.

        :return: The created project object

        API docs: https://developer.atlassian.com/bitbucket/api/2/reference/resource/workspaces/%7Bworkspace%7D/projects#post
        """
        data = {
            "name": name,
            "key": key,
            "description": description,
            "is_private": is_private,
        }
        return self.__get_object(self.post(None, data=data))

    def each(self, q=None, sort=None):
        """
        Get all projects in the workspace matching the criteria.

        :param q: string (default is None):    Query string to narrow down the response.
                                               See https://developer.atlassian.com/bitbucket/api/2/reference/meta/filtering for details.
        :param sort: string (default is None): Name of a response property to sort results.
                                               See https://developer.atlassian.com/bitbucket/api/2/reference/meta/filtering for details.

        :return: A generator for the project objects

        API docs: https://developer.atlassian.com/bitbucket/api/2/reference/resource/workspaces/%7Bworkspace%7D/projects#get
        """
        params = {}
        if sort is not None:
            params["sort"] = sort
        if q is not None:
            params["q"] = q
        for project in self._get_paged(None, params=params):
            yield self.__get_object(project)

        return

    def get(self, project, by="key"):
        """
        Returns the requested project

        :param project: string: The requested project.
        :param by: string (default is "key"): How to interpret project, can be 'key' or 'name'.

        :return: The requested Project object

        API docs: https://developer.atlassian.com/bitbucket/api/2/reference/resource/workspaces/%7Bworkspace%7D/projects/%7Bproject_key%7D#get
        """
        if by == "key":
            return self.__get_object(super(Projects, self).get(project))
        elif by == "name":
            for p in self.each():
                if p.name == project:
                    return p
        else:
            ValueError("Unknown value '{}' for argument [by], expected 'key' or 'name'".format(by))

        raise Exception("Unknown project {} '{}'".format(by, project))

    def exists(self, project, by="key"):
        """
        Check if project exist.

        :param project: string: The requested project.
        :param by: string (default is "key"): How to interpret project, can be 'key' or 'name'.

        :return: True if the project exists

        API docs: https://developer.atlassian.com/bitbucket/api/2/reference/resource/workspaces/%7Bworkspace%7D/projects/%7Bproject_key%7D#get
        """
        exists = False
        try:
            self.get(project, by)
            exists = True
        except HTTPError as e:
            if e.response.status_code in (401, 404):
                pass
        except Exception as e:
            if not str(e) == "Unknown project {} '{}'".format(by, project):
                raise e
        return exists


class Project(BitbucketCloudBase):
    def __init__(self, data, *args, **kwargs):
        super(Project, self).__init__(None, *args, data=data, expected_type="project", **kwargs)
        try:
            url = self.get_link("repositories")
        except KeyError:
            workspace = self.get_data("workspace")
            url = '{}/?q=project.key="{}"'.format(workspace["links"]["self"], workspace["slug"])
        self.__repositories = ProjectRepositories(url, **self._new_session_args)

    def update(self, **kwargs):
        """
        Update the project properties. Fields not present in the request body are ignored.

        :param kwargs: dict: The data to update.

        :return: The updated project

        API docs: https://developer.atlassian.com/bitbucket/api/2/reference/resource/workspaces/%7Bworkspace%7D/projects/%7Bproject_key%7D#put
        """
        return self._update_data(self.put(None, data=kwargs))

    def delete(self):
        """
        Delete the project.

        :return: The response on success

        API docs: https://developer.atlassian.com/bitbucket/api/2/reference/resource/workspaces/%7Bworkspace%7D/projects/%7Bproject_key%7D#delete
        """
        data = super(Project, self).delete(None)
        if data is None or "errors" in data:
            return
        return data

    @property
    def name(self):
        """The project name"""
        return self.get_data("name")

    @name.setter
    def name(self, name):
        """Setter for the project name"""
        return self.update(name=name)

    @property
    def key(self):
        """The project key"""
        return self.get_data("key")

    @key.setter
    def key(self, key):
        """Setter for the project key"""
        return self.update(key=key)

    @property
    def description(self):
        """The project description"""
        return self.get_data("description")

    @description.setter
    def description(self, description):
        """Setter for the project description"""
        return self.update(description=description)

    @property
    def is_private(self):
        """The project private flag"""
        return self.get_data("is_private")

    @is_private.setter
    def is_private(self, is_private):
        """Setter for the project private flag"""
        return self.update(is_private=is_private)

    @property
    def created_on(self):
        """The project creation time"""
        return self.get_data("created_on")

    @property
    def updated_on(self):
        """The project last update time"""
        return self.get_data("updated_on", "never updated")

    def get_avatar(self):
        """The project avatar"""
        return self.get(self.get_link("avatar"), absolute=True)

    @property
    def repositories(self):
        """The project repositories"""
        return self.__repositories



# coding: utf8
import pytest
import sys

from atlassian import ServiceDesk

SERVICEDESK = None
try:
    from .mockup import mockup_server

    SERVICEDESK = ServiceDesk(
        "{}/servicedesk".format(mockup_server()), username="username", password="password", cloud=True
    )
except ImportError:
    pass


@pytest.mark.skipif(sys.version_info < (3, 4), reason="requires python3.4")
class TestBasic:
    def test_create_customer(self):
        result = SERVICEDESK.create_customer("{displayName}", "{emailAddress}")["emailAddress"]
        assert result == "{emailAddress}", "Result of [create_customer()]"

    def test_get_organisations(self):
        result = [x["name"] for x in SERVICEDESK.get_organisations()["values"]]
        assert result == ["Charlie Cakes Franchises"], "Result of [get_organisations()]"

    def test_get_organisations_servicedesk_id(self):
        result = [x["name"] for x in SERVICEDESK.get_organisations(service_desk_id="{serviceDeskId}")["values"]]
        assert result == [
            "Charlie Cakes Franchises",
            "Atlas Coffee Co",
            "The Adjustment Bureau",
        ], "Result of [get_organisations(service_desk_id)]"

    def test_get_organization(self):
        result = SERVICEDESK.get_organization("{organizationId}")
        assert result["name"] == "Charlie Cakes Franchises", "Result of [test_get_organization(...)]"

    def test_get_users_in_organization(self):
        result = [x["emailAddress"] for x in SERVICEDESK.get_users_in_organization("{organizationId}")["values"]]
        assert result == ["fred@example.com", "bob@example.com"], "Result of [get_users_in_organization(...)]"

    def test_create_organization(self):
        result = SERVICEDESK.create_organization("Charlie Chocolate Franchises")
        assert result["id"] == "2", "Result of [create_organization(...)]"

    def test_add_organization(self):
        result = SERVICEDESK.add_organization("{serviceDeskId}", "{organizationId}")
        assert result is None, "Result of [add_organization(...)]"

    def test_remove_organization(self):
        result = SERVICEDESK.remove_organization("{serviceDeskId}", "{organizationId}")
        assert result is None, "Result of [remove_organization(...)]"

    def test_delete_organization(self):
        result = SERVICEDESK.delete_organization("{organizationId}")
        assert result is None, "Result of [delete_organization(...)]"

    def test_add_users_to_organization(self):
        result = SERVICEDESK.add_users_to_organization(
            "{organizationId}", account_list=["{accountId1}", "{accountId2}"]
        )
        assert result is None, "Result of [add_users_to_organization(...)]"

    def test_remove_users_from_organization(self):
        result = SERVICEDESK.remove_users_from_organization(
            "{organizationId}", account_list=["{accountId1}", "{accountId2}"]
        )
        assert result is None, "Result of [remove_users_from_organization(...)]"

    def test_get_customers(self):
        result = [x["emailAddress"] for x in SERVICEDESK.get_customers("{serviceDeskId}", query=None)["values"]]
        assert result == ["fred@example.com"], "Result of [remove_users_from_organization(...)]"

    def test_add_customers(self):
        result = SERVICEDESK.add_customers("{serviceDeskId}", list_of_accountids=["{accountId1}", "{accountId2}"])
        assert result is None, "Result of [remove_users_from_organization(...)]"

    def test_remove_customers(self):
        result = SERVICEDESK.remove_customers("{serviceDeskId}", list_of_accountids=["{accountId1}", "{accountId2}"])
        assert result is None, "Result of [remove_users_from_organization(...)]"



# coding=utf-8

from .base import BitbucketCloudBase
from .workspaces import Workspaces
from .repositories import Repositories


class Cloud(BitbucketCloudBase):
    def __init__(self, url="https://api.bitbucket.org/", *args, **kwargs):
        kwargs["cloud"] = True
        kwargs["api_root"] = None
        kwargs["api_version"] = "2.0"
        url = url.strip("/") + "/{}".format(kwargs["api_version"])
        super(Cloud, self).__init__(url, *args, **kwargs)
        self.__workspaces = Workspaces("{}/workspaces".format(self.url), **self._new_session_args)
        self.__repositories = Repositories("{}/repositories".format(self.url), **self._new_session_args)

    @property
    def workspaces(self):
        return self.__workspaces

    @property
    def repositories(self):
        return self.__repositories



# coding: utf8
from atlassian.bitbucket.cloud.repositories import WorkspaceRepositories
import pytest
import sys
from datetime import datetime

from atlassian import Bitbucket
from atlassian.bitbucket import Cloud
from atlassian.bitbucket.cloud.common.users import User
from atlassian.bitbucket.cloud.repositories.pullRequests import Comment, Participant, PullRequest, Build

BITBUCKET = None
try:
    from .mockup import mockup_server

    BITBUCKET = Bitbucket(
        "{}/bitbucket/cloud".format(mockup_server()), username="username", password="password", cloud=True
    )
    CLOUD = Cloud("{}/bitbucket/cloud".format(mockup_server()), username="username", password="password")
except ImportError:
    pass


def _datetimetostr(dtime):
    # convert datetime object to str because datetime.timezone is not available in py27
    # doesn't work on py27: datetime(2020, 12, 27, 14, 9, 14, 660262, tzinfo=timezone.utc)
    return dtime.strftime("%Y-%m-%d %H:%M:%S.%f")


@pytest.mark.skipif(sys.version_info < (3, 4), reason="requires python3.4")
class TestBasic:
    def test_exists_workspace(self):
        assert CLOUD.workspaces.exists("TestWorkspace1"), "Exists workspace"

    def test_not_exists_workspace(self):
        assert not CLOUD.workspaces.exists("TestWorkspace1xxx"), "Not exists workspace"

    def test_not_exists_project(self):
        assert not CLOUD.workspaces.get("TestWorkspace1").projects.exists("pr1"), "Not exists repository"

    def test_exists_repository(self):
        assert CLOUD.workspaces.get("TestWorkspace1").repositories.exists("testrepository1"), "Exists repository"

    def test_not_exists_repository(self):
        assert not CLOUD.workspaces.get("TestWorkspace1").repositories.exists(
            "testrepository1xxx"
        ), "Not exists repository"

    def test_create_repositories(self):
        result = CLOUD.workspaces.get("TestWorkspace1").repositories.create("TestRepositoryCreated")
        assert result.description == "Test newly created repository", "Result of repositories [create(...)]"
        result = CLOUD.workspaces.get("TestWorkspace1").repositories.create(
            "TestRepositoryCreated", "TEST1", fork_policy=WorkspaceRepositories.ALLOW_FORKS
        )
        assert result.fork_policy == "allow_forks", "Result of repositories [create(...)] with project and fork policy"

    def test_get_repositories(self):
        result = [x["name"] for x in BITBUCKET.get_repositories("TestWorkspace1")]
        assert result == ["testrepository1", "testrepository2"], "Result of [get_repositories(...)]"

    def test_get_pipelines(self):
        result = [x["uuid"] for x in BITBUCKET.get_pipelines("TestWorkspace1", "testrepository1")]
        assert result == ["{PipelineUuid}"], "Result of [get_pipelines(...)]"

    def test_trigger_pipeline(self):
        result = BITBUCKET.trigger_pipeline("TestWorkspace1", "testrepository1")
        assert result["uuid"] == "{PipelineUuid}", "Result of [trigger_pipeline(...)]"

    def test_get_pipeline(self):
        result = BITBUCKET.get_pipeline("TestWorkspace1", "testrepository1", "{PipelineUuid}")
        assert result["state"]["name"] == "COMPLETED", "Result of [get_pipeline(...)]"
        result = (
            CLOUD.workspaces.get("TestWorkspace1").repositories.get("testrepository1").pipelines.get("{PipelineUuid}")
        )
        assert result.get_data("state")["name"] == "COMPLETED", "Pipeline state"
        assert result.completed_on is None, "Pipeline completed time"

    def test_stop_pipeline(self):
        result = BITBUCKET.stop_pipeline("TestWorkspace1", "testrepository1", "{PipelineUuid}")
        assert result == {}, "Result of [stop_pipeline(...)]"

    def test_get_pipeline_steps(self):
        result = [
            x["uuid"] for x in BITBUCKET.get_pipeline_steps("TestWorkspace1", "testrepository1", "{PipelineUuid}")
        ]
        assert result == ["{PipelineStep1Uuid}", "{PipelineStep2Uuid}"], "Result of [get_pipeline_steps(...)]"

    def test_get_pipeline_step(self):
        result = BITBUCKET.get_pipeline_step(
            "TestWorkspace1", "testrepository1", "{PipelineUuid}", "{PipelineStep1Uuid}"
        )
        assert result["uuid"] == "{PipelineStep1Uuid}", "Result of [get_pipeline_step(...)]"

    def test_get_pipeline_step_log_1(self):
        result = BITBUCKET.get_pipeline_step_log(
            "TestWorkspace1", "testrepository1", "{PipelineUuid}", "{PipelineStep1Uuid}"
        )
        assert result is None, "Result of step1 [get_pipeline_step_log(...)]"

    def test_get_pipeline_step_log_2(self):
        result = BITBUCKET.get_pipeline_step_log(
            "TestWorkspace1", "testrepository1", "{PipelineUuid}", "{PipelineStep2Uuid}"
        )
        assert result == b"Log content", "Result of step2 [get_pipeline_step_log(...)]"

    def test_get_issues(self):
        result = [x["title"] for x in BITBUCKET.get_issues("TestWorkspace1", "testrepository1")]
        assert result == ["First issue", "Second issue"], "Result of [get_issues(...)]"

    def test_create_issue(self):
        result = BITBUCKET.create_issue("TestWorkspace1", "testrepository1", "Title", "Description", "bug", "minor")[
            "content"
        ]["raw"]
        assert result == "Description", "Result of [create_issue(...)]"

    def test_get_issue(self):
        result = BITBUCKET.get_issue("TestWorkspace1", "testrepository1", 3)["kind"]
        assert result == "bug", "Result of [get_issue(...)]"

    def test_update_issue(self):
        result = BITBUCKET.update_issue("TestWorkspace1", "testrepository1", 3, kind="enhancement")["kind"]
        assert result == "enhancement", "Result of [update_issue(...)]"

    def test_delete_issue(self):
        result = BITBUCKET.delete_issue("TestWorkspace1", "testrepository1", 3)["title"]
        assert result == "Title deleted issue", "Result of [get_issue(...)]"

    def test_get_branch_restrictions(self):
        result = [x["kind"] for x in BITBUCKET.get_branch_restrictions("TestWorkspace1", "testrepository1")]
        assert result == [
            "delete",
            "force",
            "delete",
            "restrict_merges",
            "push",
        ], "Result of [get_branch_restrictions(...)]"

    def test_update_branch_restriction(self):
        result = BITBUCKET.update_branch_restriction("TestWorkspace1", "testrepository1", 17203842, branch="master")[
            "pattern"
        ]
        assert result == "master", "Result of [update_branch_restrictions(...)]"

    def test_delete_branch_restriction(self):
        result = BITBUCKET.delete_branch_restriction("TestWorkspace1", "testrepository1", 17203842)["pattern"]
        assert result == "deleted_branch", "Result of [update_branch_restrictions(...)]"

    def test_get_default_reviewers(self):
        result = [x["display_name"] for x in BITBUCKET.get_default_reviewers("TestWorkspace1", "testrepository1")]
        assert result == ["DefaultReviewer1"], "Result of [get_default_reviewers(...)]"

    def test_is_default_reviewer(self):
        result = BITBUCKET.is_default_reviewer("TestWorkspace1", "testrepository1", "DefaultReviewerNo")
        assert result is False, "Result of [is_default_reviewer(...)]"
        result = BITBUCKET.is_default_reviewer("TestWorkspace1", "testrepository1", "DefaultReviewer1Uuid")
        assert result is True, "Result of [is_default_reviewer(...)]"

    def test_delete_default_reviewer(self):
        result = BITBUCKET.delete_default_reviewer("TestWorkspace1", "testrepository1", "DefaultReviewer1Uuid")[
            "account_id"
        ]
        assert result == "DefaultReviewer1AccountId", "Result of [delete_default_reviewer(...)]"


@pytest.mark.skipif(sys.version_info < (3, 4), reason="requires python3.4")
class TestPullRequests:
    @pytest.fixture(scope="module")
    def tc1(self):
        return CLOUD.workspaces.get("TestWorkspace1").repositories.get("testrepository1").pullrequests.get(1)

    @pytest.fixture(scope="module")
    def tc2(self):
        return CLOUD.workspaces.get("TestWorkspace1").repositories.get("testrepository1").pullrequests

    def test_id(self, tc1):
        assert tc1.id == 1

    def test_title(self, tc1):
        assert tc1.title == "PRTitle"

    def test_description(self, tc1):
        assert tc1.description == "PRDescription"

    def test_is_declined(self, tc1):
        assert not tc1.is_declined

    def test_is_merged(self, tc1):
        assert not tc1.is_merged

    def test_is_open(self, tc1):
        assert tc1.is_open

    def test_is_superseded(self, tc1):
        assert not tc1.is_superseded

    def test_created_on(self, tc1):
        assert _datetimetostr(tc1.created_on) == _datetimetostr(datetime(2020, 3, 19, 12, 0, 3, 494356))

    def test_updated_on(self, tc1):
        assert _datetimetostr(tc1.updated_on) == _datetimetostr(datetime(2020, 12, 27, 14, 9, 14, 660262))

    def test_close_source_branch(self, tc1):
        assert tc1.close_source_branch

    def test_source_branch(self, tc1):
        assert tc1.source_branch == "feature/test-branch"

    def test_destination_branch(self, tc1):
        assert tc1.destination_branch == "master"

    def test_comment_count(self, tc1):
        assert tc1.comment_count == 5

    def test_task_count(self, tc1):
        assert tc1.task_count == 0

    def test_declined_reason(self, tc1):
        assert not tc1.declined_reason

    def test_author(self, tc1):
        assert isinstance(tc1.author, User)
        assert tc1.author.display_name == "User03DisplayName"
        assert tc1.author.uuid == "{User03UUID}"
        assert tc1.author.account_id == "User03AccountID"
        assert tc1.author.nickname == "User03Nickname"

    def test_participants(self, tc1):
        participants = list(tc1.participants())
        assert len(participants) == 5

        p1 = participants[1]
        assert isinstance(p1, Participant)
        assert isinstance(p1.user, User)
        assert p1.user.display_name == "User03DisplayName"
        assert p1.user.uuid == "{User03UUID}"
        assert p1.user.account_id == "User03AccountID"
        assert p1.user.nickname == "User03Nickname"
        assert _datetimetostr(p1.participated_on) == _datetimetostr(datetime(2020, 7, 9, 7, 0, 54, 416331))
        assert p1.is_participant
        assert not p1.is_reviewer
        assert not p1.has_approved
        assert not p1.has_changes_requested

        p2 = participants[2]
        assert p2.has_approved
        assert p2.is_reviewer
        assert not p2.is_participant
        assert not p2.has_changes_requested

        p3 = participants[3]
        assert not p3.has_approved
        assert p3.is_reviewer
        assert not p3.is_participant
        assert p3.has_changes_requested

    def test_reviewers(self, tc1):
        reviewers = list(tc1.reviewers())
        assert len(reviewers) == 3
        assert isinstance(reviewers[0], User)

    def test_comment(self, tc1):
        msg = "hello world"
        com = tc1.comment(msg)
        assert com["type"] == "pullrequest_comment"
        assert com["content"]["raw"] == msg
        assert com["pullrequest"]["id"] == 1
        assert not com["deleted"]

    def test_comments(self, tc1):
        comments = list(tc1.comments())
        assert len(comments) == 2
        c1, c2 = comments
        assert isinstance(c1, Comment)
        assert isinstance(c2, Comment)
        assert c1.raw == "Test comment 1"
        assert isinstance(c1.user, User)
        assert c1.user.display_name == "User04DisplayName"
        assert c2.html == "<p>Test comment 2</p>"

    def test_approve(self, tc1):
        ap = tc1.approve()
        assert ap["approved"]
        assert ap["state"] == "approved"

    def test_unapprove(self, tc1):
        assert tc1.unapprove() is None

    def test_request_changes(self, tc1):
        rc = tc1.request_changes()
        assert rc["approved"] is False
        assert rc["state"] == Participant.CHANGES_REQUESTED

    def test_unrequest_changes(self, tc1):
        assert tc1.unrequest_changes() is None

    def test_decline(self, tc1):
        decline = tc1.decline()
        assert decline["type"] == "pullrequest"
        assert decline["state"] == PullRequest.STATE_DECLINED
        assert decline["merge_commit"] is None
        assert decline["closed_by"]["uuid"] == "{User04UUID}"

    def test_merge(self, tc1):
        merge = tc1.merge()
        assert merge["type"] == "pullrequest"
        assert merge["state"] == PullRequest.STATE_MERGED
        assert merge["closed_by"]["uuid"] == "{User04UUID}"
        assert merge["merge_commit"]["hash"] == "36bb9607a8c9e0c6222342486e3393ae154b46c0"

    def test_each(self, tc2):
        prs = list(tc2.each())
        assert len(prs) == 2
        assert isinstance(prs[0], PullRequest)
        assert prs[0].id == 1
        assert prs[1].id == 25

    def test_create(self, tc2):
        reviewers = ["{User04UUID}", "{User02UUID}", "{User01UUID}"]
        pr = tc2.create(
            title="PRTitle",
            source_branch="feature/test-branch",
            destination_branch="master",
            description="PRDescription",
            close_source_branch=True,
            reviewers=reviewers,
        )
        assert pr.id == 1
        assert len(list(pr.reviewers())) == 3

    def test_builds(self, tc1):
        builds = list(tc1.builds())
        assert len(builds) == 1

        build = builds[0]
        assert isinstance(build, Build)
        assert build.successful
        assert build.name == "Build #5"
        assert build.description == "This commit looks good."
        assert build.key == "12345"
        assert _datetimetostr(build.created_on) == _datetimetostr(datetime(2021, 1, 22, 14, 47, 16, 366063))
        assert build.commit == "1fbd047cd99a"
        assert build.refname == "feature/abranch"
        assert build.website == "http://urltothebui.ld/5"



# coding: utf8
import io
import pytest
import sys
import zipfile

from atlassian.bitbucket.server import Server

BITBUCKET = None
try:
    from .mockup import mockup_server

    BITBUCKET = Server(
        "{}/bitbucket/server".format(mockup_server()), username="username", password="password", cloud=True
    )
except ImportError:
    pass


@pytest.mark.skipif(sys.version_info < (3, 4), reason="requires python3.4")
class TestBasic:
    def test_global_permissions(self):
        result = list(BITBUCKET.groups.each())
        assert [x.name for x in result] == ["group_a", "group_b", "group_c", "group_d"], "Each global group"
        assert [x.permission for x in result] == [
            "SYS_ADMIN",
            "ADMIN",
            "PROJECT_CREATE",
            "LICENSED_USER",
        ], "The group permission"
        assert [x.is_sys_admin for x in result] == [True, False, False, False], "Is sys admin group"
        assert [x.is_admin for x in result] == [False, True, False, False], "Is admin group"
        assert [x.is_project_create for x in result] == [False, False, True, False], "Is project create group"
        assert [x.is_licensed_user for x in result] == [False, False, False, True], "Is licenced group"

        result = list(BITBUCKET.groups.each_none())
        assert [x.name for x in result] == ["group_a_1", "group_b_1"], "Each none global group"

        group = BITBUCKET.groups.get("group_a")
        assert group.name == "group_a", "Get a group"
        assert group.delete() == {}, "Delete a group"

        result = list(BITBUCKET.users.each())
        assert [x.permission for x in result] == [
            "SYS_ADMIN",
            "ADMIN",
            "PROJECT_CREATE",
            "LICENSED_USER",
        ], "The user permission"
        assert [x.name for x in result] == ["jcitizen1", "jcitizen2", "jcitizen3", "jcitizen4"], "Each global user"
        assert [x.email for x in result] == [
            "jane1@example.com",
            "jane2@example.com",
            "jane3@example.com",
            "jane4@example.com",
        ], "Each global user email"
        assert [x.displayname for x in result] == [
            "Jane Citizen 1",
            "Jane Citizen 2",
            "Jane Citizen 3",
            "Jane Citizen 4",
        ], "Each global user display name"
        assert [x.active for x in result] == [True, True, False, False], "Each global user active flag"
        assert [x.slug for x in result] == [
            "jcitizen1_slug",
            "jcitizen2_slug",
            "jcitizen3_slug",
            "jcitizen4_slug",
        ], "Each global user slug"
        assert [x.id for x in result] == [101, 102, 103, 104], "Each global user id"
        assert [x.is_sys_admin for x in result] == [True, False, False, False], "Is sys admin user"
        assert [x.is_admin for x in result] == [False, True, False, False], "Is admin user"
        assert [x.is_project_create for x in result] == [False, False, True, False], "Is project create user"
        assert [x.is_licensed_user for x in result] == [False, False, False, True], "Is licensed user"

        result = list(BITBUCKET.users.each_none())
        assert [x.name for x in result] == ["jcitizen1_1"], "Each none global user"

        user = BITBUCKET.users.get("jcitizen1")
        assert user.name == "jcitizen1", "Get a user"
        assert user.delete() == {}, "Delete a user"

    def test_projects(self):
        assert not BITBUCKET.projects.exists("PRJxxx"), "Not exists project by key"
        assert BITBUCKET.projects.exists("PRJ"), "Exists project by key"
        result = list(BITBUCKET.projects.each())
        assert [x.key for x in result] == ["PRJ", "PRJ1"], "Each project keys"
        assert [x.description for x in result] == [
            "The description for my cool project.",
            "The description for my cool project 1.",
        ], "Each project description"

        assert not BITBUCKET.projects.exists("PRJxxx"), "Not exists project by key"
        assert BITBUCKET.projects.exists("PRJ"), "Exists project by key"
        project = BITBUCKET.projects.get("PRJ")
        assert project.name == "My Cool Project", "Get project by key"
        assert not BITBUCKET.projects.exists("My Cool Project xxx", by="name"), "Not exists project by name"
        assert BITBUCKET.projects.exists("My Cool Project", by="name"), "Exists project by name"
        project = BITBUCKET.projects.get("My Cool Project", by="name")
        assert project.key == "PRJ", "Get project by name"

        assert project.id == 1, "The project id"
        assert project.type == "NORMAL", "The project type"
        assert project.name == "My Cool Project", "The project name"
        project.name = "New name"
        assert project.name == "New name", "Update the project name"

        assert project.description == "The description for my cool project.", "The project description"
        project.description = "New description."
        assert project.description == "New description.", "Update the project description"

        assert project.public is True, "The project public flag"
        project.public = False
        assert project.public is False, "Update the project public flag"

        assert project.key == "PRJ", "The project key"
        project.key = "NEWKEY"
        assert project.key == "NEWKEY", "Update the project key"

    def test_project_permissions(self):
        project = BITBUCKET.projects.get("PRJ")

        result = list(project.groups.each())
        assert [x.permission for x in result] == ["ADMIN", "WRITE", "READ"], "Each permission of project group"
        assert [x.name for x in result] == ["group_a", "group_b", "group_c"], "Each project group name"
        assert [x.is_admin for x in result] == [True, False, False], "Is admin group"
        assert [x.is_write for x in result] == [False, True, False], "Is write group"
        assert [x.is_read for x in result] == [False, False, True], "Is read group"
        assert [x.can_write for x in result] == [True, True, False], "Can group write"

        result = list(project.groups.each_none())
        assert [x.name for x in result] == ["group_a_1", "group_b_1"], "Each none project group"

        group = project.groups.get("group_a")
        assert group.name == "group_a", "Get a group"
        assert group.delete() == {}, "Delete a group"

        result = list(project.users.each())
        assert [x.permission for x in result] == ["ADMIN", "WRITE", "READ"], "Each permission of project user"
        assert [x.name for x in result] == ["jcitizen1", "jcitizen2", "jcitizen3"], "Each project user name"
        assert [x.email for x in result] == [
            "jane1@example.com",
            "jane2@example.com",
            "jane3@example.com",
        ], "Each project user email"
        assert [x.displayname for x in result] == [
            "Jane Citizen 1",
            "Jane Citizen 2",
            "Jane Citizen 3",
        ], "Each project user display name"
        assert [x.active for x in result] == [True, True, False], "Each project user active flag"
        assert [x.slug for x in result] == [
            "jcitizen1_slug",
            "jcitizen2_slug",
            "jcitizen3_slug",
        ], "Each project user slug"
        assert [x.id for x in result] == [101, 102, 103], "Each project user id"
        assert [x.is_admin for x in result] == [True, False, False], "Is admin user"
        assert [x.is_write for x in result] == [False, True, False], "Is write user"
        assert [x.is_read for x in result] == [False, False, True], "Is read user"
        assert [x.can_write for x in result] == [True, True, False], "Can user write"

        result = list(project.users.each_none())
        assert [x.name for x in result] == ["jcitizen1_1"], "Each none project user"

        user = project.users.get("jcitizen1")
        assert user.name == "jcitizen1", "Get a user"
        assert user.delete() == {}, "Delete a user"

    def test_repositories(self):
        project = BITBUCKET.projects.get("PRJ")
        result = list(project.repos.each())
        assert [x.name for x in result] == ["My repo 1", "My repo 2", "My repo 3"], "Each repo name"
        assert [x.description for x in result] == [
            "My repo 1 description",
            "My repo 2 description",
            "My repo 3 description",
        ], "Each repo description"

        print(list(project.repos.each()))

        assert project.repos.exists("my-repo1-slug"), "Repo exists by slug"

        repo = project.repos.get("my-repo1-slug")
        assert repo.name == "My repo 1", "The repo name by slug"
        assert repo.id == 1, "The repo id"

        repo = project.repos.get("My repo 1", by="name")
        assert repo.name == "My repo 1", "The repo name"
        repo.name = "New name"
        assert repo.name == "New name", "Update the repo name"

        assert repo.slug == "my-repo1-slug", "The repo slug"

        assert repo.description == "My repo 1 description", "The repo description"
        repo.description = "New description."
        assert repo.description == "New description.", "Update the repo description"

        assert repo.public is True, "The repo public flag"
        repo.public = False
        assert repo.public is False, "Update the repo public flag"

        assert repo.forkable is True, "The repo forkable flag"
        repo.forkable = False
        assert repo.forkable is False, "Update the repo forkable flag"

        assert repo.contributing() == "Test contributing.md", "The contributing.md"
        assert repo.contributing(at="CommitId") == "Test contributing.md at CommitId", "The contributing.md at CommitId"
        assert (
            repo.contributing(at="CommitId", markup=True) == "<p>Test rendered contributing.md at CommitId</p>"
        ), "The rendered contributing.md at CommitId"

        assert repo.license() == "Test license.md", "The license.md"
        assert repo.license(at="CommitId") == "Test license.md at CommitId", "The license.md at CommitId"
        assert (
            repo.license(at="CommitId", markup=True) == "<p>Test rendered license.md at CommitId</p>"
        ), "The rendered license.md at CommitId"

        assert repo.readme() == "Test readme.md", "The readme.md"
        assert repo.readme(at="CommitId") == "Test readme.md at CommitId", "The readme.md at CommitId"
        assert (
            repo.readme(at="CommitId", markup=True) == "<p>Test rendered readme.md at CommitId</p>"
        ), "The rendered readme.md at CommitId"

        assert repo.default_branch == "main", "The default branch"
        repo.default_branch = "maint/relx"

        assert [x["hierarchyId"] for x in repo.forks()] == ["e3c939f9ef4a7fae272e"], "The forks"

        assert [x.id for x in repo.related()] == [2], "The related repositories"

    def test_repository_permissions(self):
        repo = BITBUCKET.projects.get("PRJ").repos.get("my-repo1-slug")

        result = list(repo.groups.each())
        assert [x.permission for x in result] == ["ADMIN", "WRITE", "READ"], "Each permission of repo group"
        assert [x.name for x in result] == ["group_a", "group_b", "group_c"], "Each repo group name"
        assert [x.is_admin for x in result] == [True, False, False], "Is admin group"
        assert [x.is_write for x in result] == [False, True, False], "Is write group"
        assert [x.is_read for x in result] == [False, False, True], "Is read group"
        assert [x.can_write for x in result] == [True, True, False], "Can group write"

        result = list(repo.groups.each_none())
        assert [x.name for x in result] == ["group_a_1", "group_b_1"], "Each none repo group"

        group = repo.groups.get("group_a")
        assert group.name == "group_a", "Get a group"
        assert group.delete() == {}, "Delete a group"

        result = list(repo.users.each())
        assert [x.permission for x in result] == ["ADMIN", "WRITE", "READ"], "Each permission of repo user"
        assert [x.name for x in result] == ["jcitizen1", "jcitizen2", "jcitizen3"], "Each repo user name"
        assert [x.email for x in result] == [
            "jane1@example.com",
            "jane2@example.com",
            "jane3@example.com",
        ], "Each repo user email"
        assert [x.displayname for x in result] == [
            "Jane Citizen 1",
            "Jane Citizen 2",
            "Jane Citizen 3",
        ], "Each repo user display name"
        assert [x.active for x in result] == [True, True, False], "Each repo user active flag"
        assert [x.slug for x in result] == ["jcitizen1_slug", "jcitizen2_slug", "jcitizen3_slug"], "Each repo user slug"
        assert [x.id for x in result] == [101, 102, 103], "Each repo user id"
        assert [x.is_admin for x in result] == [True, False, False], "Is admin user"
        assert [x.is_write for x in result] == [False, True, False], "Is write user"
        assert [x.is_read for x in result] == [False, False, True], "Is read user"
        assert [x.can_write for x in result] == [True, True, False], "Can user write"

        result = list(repo.users.each_none())
        assert [x.name for x in result] == ["jcitizen1_1"], "Each none repo user"

        user = repo.users.get("jcitizen1")
        assert user.name == "jcitizen1", "Get a user"
        assert user.delete() == {}, "Delete a user"

    def test_download_repo_archive(self):
        repo = BITBUCKET.projects.get("PRJ").repos.get("my-repo1-slug")
        with io.BytesIO() as buf:
            repo.download_archive(buf)
            with zipfile.ZipFile(buf) as archive:
                assert archive.namelist() == ["readme.md"]
                with archive.open("readme.md") as file_in_archive:
                    assert file_in_archive.read() == b"Test readme.md"

        with io.BytesIO() as buf:
            repo.download_archive(buf, at="CommitId")
            with zipfile.ZipFile(buf) as archive:
                assert archive.namelist() == ["readme.md"]
                with archive.open("readme.md") as file_in_archive:
                    assert file_in_archive.read() == b"Test readme.md at CommitId"



# coding=utf-8

from requests import HTTPError
from .repos import Repositories
from ..base import BitbucketServerBase
from ..common.permissions import Groups, Users


class Projects(BitbucketServerBase):
    def __init__(self, url, *args, **kwargs):
        super(Projects, self).__init__(url, *args, **kwargs)

    def __get_object(self, data):
        return Project(data, **self._new_session_args)

    def create(self, name, key, description, avatar=None):
        """
        Creates a new project with the given values.

        Note that the avatar has to be embedded as either a data-url or a URL to an external image as shown in
        the examples below:

            w.projects.create( "Mars Project", "MARS", "Software for colonizing mars.",
                avatar="data:image/gif;base64,R0lGODlhEAAQAMQAAORHHOVSKudfOulrSOp3WOyDZu6QdvCchPGolfO0o/..."
            )

            w.projects.create( "Mars Project", "MARS", "Software for colonizing mars.",
                avatar="http://i.imgur.com/72tRx4w.gif"
            )

        :param name: string: The name of the project.
        :param key: string: The key of the project.
        :param description: string: The description of the project.
        :param avatar: string: The avatar of the project.

        :return: The created project object

        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp148
        """
        return self.__get_object(self.post(None, data={"name": name, "key": key, "description": description}))

    def each(self, name=None, permission=None):
        """
        Get all projects matching the criteria.

        :param name: string: Name to filter by.
        :param permission: string: Permission to filter by.

        :return: A generator for the project objects

        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp149
        """
        params = {}
        if name is not None:
            params["name"] = name
        if permission is not None:
            params["permission"] = permission
        for project in self._get_paged(None, params=params):
            yield self.__get_object(project)

        return

    def get(self, project, by="key"):
        """
        Returns the requested project

        :param project: string: The requested project.
        :param by: string (default is "key"): How to interpret project, can be 'key' or 'name'.

        :return: The requested Project object

        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp153
        """
        if by == "key":
            return self.__get_object(super(Projects, self).get(project))
        elif by == "name":
            for p in self.each(name=project):
                if p.name == project:
                    return p
        else:
            ValueError("Unknown value '{}' for argument [by], expected 'key' or 'name'".format(by))

        raise Exception("Unknown project {} '{}'".format(by, project))

    def exists(self, project, by="key"):
        """
        Check if project exist.

        :param project: string: The requested project.
        :param by: string (default is "key"): How to interpret project, can be 'key' or 'name'.

        :return: True if the project exists
        """
        exists = False
        try:
            self.get(project, by)
            exists = True
        except HTTPError as e:
            if e.response.status_code in (401, 404):
                pass
        except Exception as e:
            if not str(e) == "Unknown project {} '{}'".format(by, project):
                raise e
        return exists


class Project(BitbucketServerBase):
    def __init__(self, data, *args, **kwargs):
        super(Project, self).__init__(None, *args, data=data, **kwargs)
        self.__groups = Groups(self._sub_url("permissions/groups"), "PROJECT", **self._new_session_args)
        self.__users = Users(self._sub_url("permissions/users"), "PROJECT", **self._new_session_args)
        self.__repos = Repositories(self._sub_url("repos"), **self._new_session_args)

    def delete(self):
        """
        Delete the project.

        :return: The response on success

        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp151
        """
        return super(Project, self).delete(None)

    def update(self, **kwargs):
        """
        Update the project properties. Fields not present in the request body are ignored.

        :param kwargs: dict: The data to update.

        :return: The updated project

        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp152
        """
        return self._update_data(self.put(None, data=kwargs))

    @property
    def id(self):
        """The project identifier"""
        return self.get_data("id")

    @property
    def type(self):
        """The project type"""
        return self.get_data("type")

    @property
    def name(self):
        """The project name"""
        return self.get_data("name")

    @name.setter
    def name(self, name):
        """Setter for the project name"""
        return self.update(name=name)

    @property
    def key(self):
        """The project key"""
        return self.get_data("key")

    @key.setter
    def key(self, key):
        """Setter for the project key"""
        return self.update(key=key)

    @property
    def description(self):
        """The project description"""
        return self.get_data("description")

    @description.setter
    def description(self, description):
        """Setter for the project description"""
        return self.update(description=description)

    @property
    def public(self):
        """The project public flag"""
        return self.get_data("public")

    @public.setter
    def public(self, public):
        """Setter for the project public flag"""
        return self.update(public=public)

    def get_avatar(self, s=0):
        """
        Get the avatar.

        :param s: int (default 0): The desired size of the image. The server will return
                                   an image as close as possible to the specified size.

        :return: The response on success

        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp155
        """
        return self.get("avatar.png", params={"s": s})

    def set_avatar(self, avatar):
        """
        Set the avatar.

        :param avatar: See function create for examples.

        :return: The response on success

        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp156
        """
        return self.post("avatar.png", data={"avatar": avatar})

    @property
    def groups(self):
        """
        Property to access the project groups
        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp158
        """
        return self.__groups

    @property
    def users(self):
        """
        Property to access the project groups
        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp164
        """
        return self.__users

    @property
    def repos(self):
        """
        Property to access the repositories
        API docs: https://docs.atlassian.com/bitbucket-server/rest/7.8.0/bitbucket-rest.html#idp173
        """
        return self.__repos
